@@ -9,8 +9,6 @@
  * COPYING file.
 **/
 
-in_module(null);
-
 define_variable("active_img_hint_background_color", "#88FF00",
     "Color for the active image hint background.");
 
@@ -30,6 +28,9 @@ define_variable("hint_digits", null,
     "and ascending.  If null, base 10 will be used with the normal "+
     "hindu-arabic numerals.");
 
+define_variable("hints_display_alt", true,
+    "Display alt text in hints.");
+
 
 /**
  * hints_enumerate is a generator of natural numbers in the base defined
@@ -146,6 +147,8 @@ hint_manager.prototype = {
             var scrollY = window.scrollY;
 
             var doc = window.document;
+            if (! doc.documentElement)
+                return;
             var res = doc.evaluate(xpath_expr, doc, xpath_lookup_namespace,
                                    Ci.nsIDOMXPathResult.UNORDERED_NODE_SNAPSHOT_TYPE,
                                    null /* existing results */);
@@ -181,6 +184,18 @@ hint_manager.prototype = {
                     rect = elem.getClientRects()[0];
                 if (!rect)
                     continue;
+                var nchildren = elem.childNodes.length;
+                if (elem instanceof Ci.nsIDOMHTMLAnchorElement &&
+                    rect.width == 0 && rect.height == 0)
+                {
+                    for (var c = 0; c < nchildren; ++c) {
+                        var cc = elem.childNodes.item(c);
+                        if (cc.getBoundingClientRect) {
+                            rect = cc.getBoundingClientRect();
+                            break;
+                        }
+                    }
+                }
                 show_text = false;
                 if (elem instanceof Ci.nsIDOMHTMLInputElement || elem instanceof Ci.nsIDOMHTMLTextAreaElement)
                     text = elem.value;
@@ -192,10 +207,10 @@ hint_manager.prototype = {
                 } else if (elem instanceof Ci.nsIDOMHTMLFrameElement) {
                     text = elem.name ? elem.name : "";
                 } else if (/^\s*$/.test(elem.textContent) &&
-                           elem.childNodes.length == 1 &&
+                           nchildren == 1 &&
                            elem.childNodes.item(0) instanceof Ci.nsIDOMHTMLImageElement) {
                     text = elem.childNodes.item(0).alt;
-                    show_text = true;
+                    show_text = hints_display_alt;
                 } else
                     text = elem.textContent;
 
@@ -381,12 +396,19 @@ hint_manager.prototype = {
             if (h.visible) {
                 h.visible = false;
                 if (h.saved_color != null) {
-                    h.elem.style.color = h.saved_color;
-                    h.elem.style.backgroundColor = h.saved_bgcolor;
+                    try {
+                        h.elem.style.color = h.saved_color;
+                        h.elem.style.backgroundColor = h.saved_bgcolor;
+                    } catch (e) { /* element may be dead */ }
                 }
-                if (h.img_hint)
-                    h.img_hint.style.display = "none";
-                h.hint.style.display = "none";
+                if (h.img_hint) {
+                    try {
+                        h.img_hint.style.display = "none";
+                    } catch (e) { /* element may be dead */ }
+                }
+                try {
+                    h.hint.style.display = "none";
+                } catch (e) { /* element may be dead */ }
             }
         }
     },
@@ -394,54 +416,84 @@ hint_manager.prototype = {
     remove: function () {
         for (var i = 0, h; h = this.hints[i]; ++i) {
             if (h.visible && h.saved_color != null) {
-                h.elem.style.color = h.saved_color;
-                h.elem.style.backgroundColor = h.saved_bgcolor;
+                try {
+                    h.elem.style.color = h.saved_color;
+                    h.elem.style.backgroundColor = h.saved_bgcolor;
+                } catch (e) { /* element may be dead */ }
             }
-            if (h.img_hint)
-                h.img_hint.parentNode.removeChild(h.img_hint);
-            h.hint.parentNode.removeChild(h.hint);
+            if (h.img_hint) {
+                try {
+                    h.img_hint.parentNode.removeChild(h.img_hint);
+                } catch (e) { /* element may be dead */ }
+            }
+            try {
+                h.hint.parentNode.removeChild(h.hint);
+            } catch (e) { /* element may be dead */ }
         }
         this.hints = [];
         this.valid_hints = [];
     }
 };
 
-/* Show panel with currently selected URL. */
-function hints_url_panel (hints, window) {
-    var g = new dom_generator(window.document, XUL_NS);
-
-    var p = g.element("hbox", "class", "panel url", "flex", "0");
-    g.element("label", p, "value", "URL:", "class", "url-panel-label");
-    var url_value = g.element("label", p, "class", "url-panel-value",
-                              "crop", "end", "flex", "1");
-    window.minibuffer.insert_before(p);
-
-    p.update = function () {
-	url_value.value = "";
-	if (hints.manager && hints.manager.last_selected_hint) {
-            var spec;
-            try {
-                spec = load_spec(hints.manager.last_selected_hint.elem);
-            } catch (e) {}
-            if (spec) {
-                var uri = load_spec_uri_string(spec);
-                if (uri) url_value.value = uri;
-            }
-	}
-    };
-
-    p.destroy = function () {
-        this.parentNode.removeChild(this);
-    };
-
-    return p;
+/**
+ * Display the URL and other information for the currently selected node.
+ */
+function hints_minibuffer_annotation (hints, window) {
+    this.hints = hints;
+    this.input = window.minibuffer.input_element;
 }
+hints_minibuffer_annotation.prototype = {
+    constructor: hints_minibuffer_annotation,
 
-define_variable("hints_display_url_panel", false,
-    "When selecting a hint, the URL can be displayed in a panel above "+
-    "the minibuffer.  This is useful for confirming that the correct "+
-    "link is selected and that the URL is not evil.  This option is "+
-    "most useful when hints_auto_exit_delay is long or disabled.");
+    update: function () {
+        var s = [];
+        if (this.hints.manager && this.hints.manager.last_selected_hint) {
+            var elem = this.hints.manager.last_selected_hint.elem;
+            if (elem.hasAttribute("onmousedown") ||
+                elem.hasAttribute("onclick"))
+            {
+                s.push("[script]");
+            }
+            var tag = elem.localName.toLowerCase();
+            if ((tag == "input" || tag == "button") &&
+                elem.type == "submit" && elem.form && elem.form.action)
+            {
+                s.push((elem.form.method || "GET").toUpperCase() + ":" +
+                       elem.form.action);
+            } else {
+                try {
+                    var spec = load_spec(elem);
+                    var uri = load_spec_uri_string(spec);
+                    if (uri)
+                        s.push(uri);
+                } catch (e) {}
+            }
+        }
+        this.input.annotation = s.join(" ");
+    },
+
+    load: function () {
+        this.input.annotate = true;
+        this.update();
+    },
+
+    unload: function () {
+        this.input.annotate = false;
+    }
+};
+
+define_global_mode("hints_minibuffer_annotation_mode",
+    function enable () {
+        minibuffer_annotation_mode.register(hints_minibuffer_annotation_mode);
+    },
+    function disable () {
+        minibuffer_annotation_mode.unregister(hints_minibuffer_annotation_mode);
+    },
+    $doc = "Display the URL associated with the currently selected hint in "+
+           "a minibuffer annotation.\nThis mode is most useful when "+
+           "hints_auto_exit_delay is long or disabled.");
+
+hints_minibuffer_annotation_mode(true);
 
 /**
  * keyword arguments:
@@ -451,14 +503,18 @@ define_variable("hints_display_url_panel", false,
  * $abort_callback
  */
 define_keywords("$keymap", "$auto", "$hint_xpath_expression", "$multiple");
-function hints_minibuffer_state (minibuffer, continuation, buffer) {
+function hints_minibuffer_state (minibuffer, buffer) {
     keywords(arguments, $keymap = hint_keymap, $auto);
     basic_minibuffer_state.call(this, minibuffer, $prompt = arguments.$prompt,
                                 $keymap = arguments.$keymap);
-    if (hints_display_url_panel)
-	this.url_panel = hints_url_panel(this, buffer.window);
+    if (hints_minibuffer_annotation_mode_enabled)
+        this.hints_minibuffer_annotation = new hints_minibuffer_annotation(this, buffer.window);
     this.original_prompt = arguments.$prompt;
-    this.continuation = continuation;
+
+    let deferred = Promise.defer();
+    this.deferred = deferred;
+    this.promise = make_simple_cancelable(deferred);
+
     this.auto_exit = arguments.$auto ? true : false;
     this.xpath_expr = arguments.$hint_xpath_expression;
     this.auto_exit_timer_ID = null;
@@ -480,8 +536,8 @@ hints_minibuffer_state.prototype = {
                                             this.focused_frame, this.focused_element);
         }
         this.manager.update_valid_hints();
-        if (this.url_panel)
-            this.url_panel.update();
+        if (this.hints_minibuffer_annotation)
+            this.hints_minibuffer_annotation.load();
     },
     clear_auto_exit_timer: function () {
         var window = this.minibuffer.window;
@@ -493,13 +549,16 @@ hints_minibuffer_state.prototype = {
     unload: function () {
         this.clear_auto_exit_timer();
         this.manager.hide_hints();
+        if (this.hints_minibuffer_annotation)
+            this.hints_minibuffer_annotation.unload();
         basic_minibuffer_state.prototype.unload.call(this);
     },
     destroy: function () {
+        this.promise.cancel();
         this.clear_auto_exit_timer();
         this.manager.remove();
-        if (this.url_panel)
-            this.url_panel.destroy();
+        if (this.hints_minibuffer_annotation)
+            this.hints_minibuffer_annotation.unload();
         basic_minibuffer_state.prototype.destroy.call(this);
     },
     update_minibuffer: function (m) {
@@ -507,8 +566,8 @@ hints_minibuffer_state.prototype = {
             m.prompt = this.original_prompt + " #" + this.typed_number;
         else
             m.prompt = this.original_prompt;
-        if (this.url_panel)
-            this.url_panel.update();
+        if (this.hints_minibuffer_annotation)
+            this.hints_minibuffer_annotation.update();
     },
 
     handle_auto_exit: function (ambiguous) {
@@ -630,7 +689,7 @@ function hints_next (window, s, count) {
             cur += vl;
         s.manager.select_hint(cur + 1);
     }
-    s.update_minibuffer(window);
+    s.update_minibuffer(window.minibuffer);
 }
 interactive("hints-next", null,
     function (I) {
@@ -650,11 +709,8 @@ function hints_exit (window, s) {
     else if (cur == 0)
         elem = window.buffers.current.top_frame;
     if (elem !== null) {
-        var c = s.continuation;
-        delete s.continuation;
+        s.deferred.resolve(elem);
         window.minibuffer.pop_state();
-        if (c)
-            c(elem);
     }
 }
 
@@ -674,10 +730,9 @@ define_keywords("$buffer");
 minibuffer.prototype.read_hinted_element = function () {
     keywords(arguments);
     var buf = arguments.$buffer;
-    var s = new hints_minibuffer_state(this, (yield CONTINUATION), buf, forward_keywords(arguments));
+    var s = new hints_minibuffer_state(this, buf, forward_keywords(arguments));
     this.push_state(s);
-    var result = yield SUSPEND;
-    yield co_return(result);
+    yield co_return(yield s.promise);
 };
 
 provide("hints");
